﻿using UnityEngine;
using System.Collections;

public class MouseManager : MonoBehaviour {

	// Use this for initialization
	void Start () {
	
	}

    public GameObject ThePrefabToSpawn;

    public LayerMask SnapPointLayerMask;

    public ComponentKeybindDialog ComponentKeybindDialog;

    public GameObject ShipRoot;
	
	// Update is called once per frame
	void Update () {
	
        // Was the mouse pressed down this frame?

        if(Input.GetMouseButtonDown(0))
        {
            CheckLeftClick();
        }

        if(Input.GetMouseButtonDown(1))
        {
            CheckRightClick();
        }
	}

    Collider DoRaycast()
    {
        Camera theCamera = Camera.main;

        Ray ray = theCamera.ScreenPointToRay( Input.mousePosition );

        RaycastHit hitInfo;

        if( Physics.Raycast( ray, out hitInfo ) )
        {
            return hitInfo.collider;
        }

        return null;
    }

    void CheckRightClick()
    {
        Collider theCollider = DoRaycast();

        if(theCollider == null)
        {
            return;
        }

        // We have right-clicked on something.  Is it a KeybindableComponent?
        // We need to check the PARENT of the object with the collider, given
        // how we have assembled our prefabs.

        GameObject shipPart = FindShipPart(theCollider);

        if(shipPart == null)
        {
            // We clicked on something that doesn't have a parent, so it's probably
            // not a valid part of our ship
            return;
        }


        KeybindableComponent kc = shipPart.GetComponent<KeybindableComponent>();

        if(kc == null)
        {
            // This object isn't keybindable.
            return;
        }

        // If we get to this point, we have right-clicked on something keybindable

        ComponentKeybindDialog.OpenDialog( kc );

    }

    GameObject FindShipPart(Collider collider)
    {
        Transform curr = collider.transform;

        while(curr != null)
        {
            if(curr.gameObject.tag == "ShipPart")
            {
                return curr.gameObject;
            }

            curr = curr.parent;
        }

        return null;
    }

    void CheckLeftClick()
    {
        // Yes, the left mouse button was pressed this frame
        // Is the mouse over a cube

        Collider theCollider = DoRaycast();

        if(theCollider == null)
        {
            return;
        }

        // Did we click on a snap point?
        // Our snap points are always going to be on the SnapPoint physics layer

        // The normal integer number value of 8 is represented thusly in binary:
        //
        //      0000000000000000001000
        //
        // When you're talking about things like "LayerMasks" "BitMasks" "Flags"
        //
        //      00000000000000100010000 (in this case "Water" and "SnapPoint" are on)
        //      Would represent something like the 4th & 8th item in the list is enabled
        //      (The layermask is zero-indexed, so layer 4 & 8 are the 5th and 9th bit.)
        //
        // Bitwise operators and comparison
        //
        //  &: This kind of compares two integers and returns a new integer where
        //     both had their bits enabled
        //      Example:
        //              10110 
        //            & 10011
        //              ----
        //              10010
        //
        //  |: This is a bitwise or
        //
        //              10110 
        //            | 10011
        //              -----
        //              10111
        //
        //
        //  <<: The left-shift operator.
        //
        //     1 << 8
        //
        //    Goes from this: 0000000000000000001
        //    To this:        0000000000100000000

        int maskForThisHitObject = 1 << theCollider.gameObject.layer;


        if( (maskForThisHitObject & SnapPointLayerMask) != 0  )
        {
            // Now let's spawn a new object

            Vector3 spawnSpot = theCollider.transform.position;

            Quaternion spawnRotation = theCollider.transform.rotation; //Quaternion.FromToRotation( Vector3.forward, hitInfo.normal );

            GameObject go = (GameObject)Instantiate( ThePrefabToSpawn, spawnSpot, spawnRotation );
            go.transform.SetParent(theCollider.transform);

            // To calc mass, get a list of all the colliders in the new part
            Collider[] cols = go.transform.GetComponentsInChildren<Collider>();
            Debug.Log("Num colliders in new object: " + cols.Length);
            float mass = 0;
            foreach(Collider col in cols)
            {
                int maskForThisCol = 1 << col.gameObject.layer;
                if( (maskForThisCol & SnapPointLayerMask) == 0  )
                {
                    // This col is NOT a snap point, so calc its mass contribution
                    float volume = col.bounds.size.x * col.bounds.size.y * col.bounds.size.z;

                    // TODO???  We could check if there's a "DensityOverride" component
                    // on this new object and therefore not assume a densite of 1 on a
                    // case by case basis

                    mass += volume;  // Assume a density of 1 per unit
                }
            }
            Debug.Log("Adding mass: " + mass);
            theCollider.transform.GetComponentInParent<Rigidbody>().mass += mass;

            // Disable the renderer on the snap point while something is attached to it
            if(theCollider.GetComponent<Renderer>() != null)
            {
                theCollider.GetComponent<Renderer>().enabled = false;
            }
            if(theCollider.GetComponent<Collider>() != null)
            {
                theCollider.GetComponent<Collider>().enabled = false;
            }
        }



    }

    void RemovePart(GameObject go)
    {
        // THIS IS JUST AN EXAMPLE -- may not work right as-is

        // Make sure to re-enable the SnapPoint we are attached to

        go.transform.parent.GetComponent<Renderer>().enabled = true;
        go.transform.parent.GetComponent<Collider>().enabled = true;

        Destroy(go);
    }

    public void SetMode_Edit()
    {
        // Show all the SnapPoint nodes
        SetSnapPointEnabled(ShipRoot.transform, true);

        // Unlock the camera controls
        Camera.main.transform.parent.SetParent(null);
    }

    public void SetMode_Flight()
    {
        // Hide all the SnapPoint nodes
        SetSnapPointEnabled(ShipRoot.transform, false);

        // Tell the camera to lock to the ship's root
        Camera.main.transform.parent.SetParent(ShipRoot.transform);
        Camera.main.transform.parent.localPosition = Vector3.zero;


    }

    void SetSnapPointEnabled( Transform t, bool setToActive )
    {
        //Debug.Log("SetSnapPointEnabled: " + t.gameObject.name);
        int maskForThisHitObject = 1 << t.gameObject.layer;

        if( (maskForThisHitObject & SnapPointLayerMask) != 0  )
        {
            // This is a snap point
            if(setToActive)
            {
                // Always activate -- just in case
                t.gameObject.SetActive(true);

            }
            else
            {
                // Only inactivate the SnapPoint if it has no children (i.e. it's on the outside and visible.)
                if(t.childCount == 0)
                {
                    t.gameObject.SetActive(false);
                    return; // Exit the function.
                }
            }
        }

            
        // Loop through all of this object's children.
        for (int i = 0; i < t.childCount; i++)
        {
            // Call function recursively
            SetSnapPointEnabled(t.GetChild(i), setToActive);
        }
    }
}
